/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>

#include "config.h"

#define DBGPrint(lev,str) do {      \
  if ((lev)<=_DBG_LEVEL_) {         \
    printf ("[dhg %d] ",lev);       \
    printf str;                     \
    fflush (stdout);                \
}} while (0);

extern int _DBG_LEVEL_;

typedef signed long long int cycle_time;
typedef double us_time;
typedef double sec_time;

// ugh, hack
extern int need_rtc_timers;
extern int need_mm_timers;

extern unsigned int duration;
extern int tracing;
extern int num_threads;
extern double TSC_MHZ;
extern struct thrd_data_t *thrd_data;
extern struct trace_rec_t *trace_recs;
extern long current_record;
extern cycle_time start_time, end_time;
extern cycle_time gap;
extern double gap_us;
extern int suppress_trace;
extern double gap_pct;

extern cycle_time trace_start;
extern cycle_time trace_end;
extern int print_cycles;
extern cycle_time polling_overhead;
extern double polling_overhead_us;

extern int gap_hist_toobig;
extern int generate_gap_hist;
extern int *gap_hist;
// FIXME -- this is somewhat arbitrary
#define NUM_GAP_HIST_BUCKETS 5000
extern int GAP_HIST_DIVISOR;

#define TRUE 1
#define FALSE 0

/*
 * number of microseconds to wait before threads start doing things;
 * this helps ensure that everyone sets priorty, creates reservations,
 * etc. before the high-priority threads soak up all CPU time
 */
#define INITIAL_DELAY 5000

enum thrd_work_t { 
  CPU = 121, 
  PERIODIC,
  CPU_PERIODIC,
  CPU_SCAN,
  CPU_SCAN_YIELD,
  CPU_YIELD,
  LAT
};

enum thrd_pri_t { 
  IDLE = 222, 
  LOW, 
  NORMAL, 
  HIGH, 
  HIGHEST, 
  RTLOW, 
  RTMED, 
  RTHIGH, 
  TIMER_PRI, 
  MAIN_PRI 
};

enum thrd_status_t { 
  INIT = 333, 
  READY, 
  GO, 
  STOP, 
  DONE 
};

#define MAXSTRLEN 512
#define LAT_SLOTS 50000
#define DEF_NUM_TRACE_RECORDS 300000
#define MAX_SPIN 500

extern int NUM_TRACE_RECORDS;

/* argh - usleep of <1-2ms usually busy-waits */
#define SMALL_SLEEP 5000

#define ASSERT(x) assert(x)

struct timer_funcs {
  void (*timed_sleep) (int, us_time);
  void (*init_timers) (void);
  void (*deinit_timers) (void);
};

#ifdef WITH_RTCTIMERS
extern const struct timer_funcs rtc_timer_funcs;
#endif

#ifdef WITH_HRTIMERS
extern const struct timer_funcs hr_timer_funcs;
#endif

#ifdef WIN32
extern const struct timer_funcs mm_timer_funcs;
#endif

extern const struct timer_funcs unix_timer_funcs;

struct thrd_data_t {
  // generic stuff
  volatile enum thrd_status_t status;
  enum thrd_work_t work; 
  enum thrd_pri_t pri;
  pthread_t thrd;
  pthread_t monitor_thrd;
  int id;
  int log;
  const struct timer_funcs *timers;
  long long int workdone;

  us_time s_time, f_time;

  int *scanmem;
  unsigned int memsize;

  // periodic thread stuff
  us_time app_period, app_amount;
  int hits, misses;

  // latency test stuff
  cycle_time *lat_slots;
  int lat_slot;

  // reservation stuff
  us_time res_period, res_amount;
  int res_soft;

  // execution trace stuff
  cycle_time last_exec;
  cycle_time exec_start;
  cycle_time thread_time;
  int reserved_record;
};

struct trace_rec_t {
  int id;
  cycle_time start, end;
};

#ifdef CPU_586
/*
 * a bit of documentation since the rdrsc function seems to often be
 * misimplemented:
 *
 * rdtsc is an instruction and "=A" (d) tells gcc that we're
 * outputting a 64-bit quantity, stored in eax:edx, and putting it
 * into variable d
 *
 * when optimization is turned on, gcc does the right thing: it
 * inlines this function and (properly) eliminates storage location d
 */
static inline cycle_time rdtsc (void)
{
  cycle_time d;
  __asm__ __volatile__ ("rdtsc" : "=A" (d) );
  return d;
}

/* these routines snagged from Wine */

static inline int my_InterlockedExchangeAdd (long *dest, long incr)
{
    long ret;
    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (incr) : "memory" );
    return ret;
}

static inline int my_InterlockedIncrement (long *dest)
{
    return my_InterlockedExchangeAdd (dest, 1) + 1;
}

#endif  /* defined CPU_586 */

#ifdef CPU_PPC

static inline cycle_time rdtsc (void)
{
  unsigned long long int ts;
  unsigned long chk;
  
  __asm__ __volatile__ ("1:     mftbu   %0;"
			"       mftb    %1;"
                        "       mftbu   %2;"
                        "       cmpw    %2, %0;"
                        "       bne     1b" :     "=r" (((unsigned long int *)&ts)[0]),
			                          "=r" (((unsigned long int *)&ts)[1]),
			                          "=r" (chk));
  return ts;
}

/* these routines snagged from Wine */

static inline int my_InterlockedExchangeAdd (long *dest, long incr)
{
    long ret = 0;
    long zero = 0;
    __asm__(
        "0:    lwarx %0, %3, %1\n"
        "      add %0, %2, %0\n"
        "      stwcx. %0, %3, %1\n"
        "      bne- 0b\n"
	"      isync\n"
        : "=&r" (ret)
        : "r"(dest), "r"(incr), "r"(zero)
        : "cr0", "memory", "r0"
    );
    return ret-incr;
}

static inline int my_InterlockedIncrement (long *dest)
{
    return my_InterlockedExchangeAdd (dest, 1) + 1;
}

#endif /* defined CPU_PPC */

static inline sec_time cycle_to_sec (cycle_time c)
{
  assert (TSC_MHZ != -1);
  return c / (sec_time)(TSC_MHZ * 1000000);
}

static inline cycle_time sec_to_cycle (sec_time s)
{
  assert (TSC_MHZ != -1);
  return (cycle_time)s * (cycle_time)(TSC_MHZ * 1000000);
}

static inline cycle_time us_to_cycle (us_time us)
{
  assert (TSC_MHZ != -1);
  return (cycle_time)us * (cycle_time)TSC_MHZ;
}

static inline us_time cycle_to_us (cycle_time c)
{
  return c / (us_time)TSC_MHZ;
}

static inline us_time sec_to_us (sec_time s)
{
  return (us_time)s * 1000000;
}

static inline sec_time us_to_sec (us_time us)
{
  return (sec_time)(us/1000000);
}

extern void *xmalloc (size_t size);

extern void set_priority (enum thrd_pri_t pri);
extern void lock_pages (void);
extern double find_mhz (void);

extern void *thrd_func (void *arg);

extern void *monitor_thrd_func (void *arg);

extern void print_all_results (void);

extern void parse_args (int argc, char *argv[]);

extern void make_res (us_time amount, 
		      us_time period,
		      int soft,
		      int which);

extern void delete_res (int which);

extern pid_t hourglass_pid;

// polymorphic, but wrong when x or y has side effects
#define mymin(x,y) ((x)<(y)?(x):(y))
#define mymax(x,y) ((x)>(y)?(x):(y))

extern void print_gap_hist (void);
