/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

double TSC_MHZ = -1;
int num_threads = 0;
unsigned int duration;
int _DBG_LEVEL_ = 0;

struct thrd_data_t *thrd_data = NULL;
struct trace_rec_t *trace_recs = NULL;
long current_record;
int tracing;
cycle_time start_time, end_time;
int need_rtc_timers = 0;
int need_mm_timers = 0;
double gap_us;
cycle_time gap;
double gap_pct;
double polling_overhead_us;
cycle_time polling_overhead;

int *gap_hist = NULL;
int generate_gap_hist = FALSE;
int gap_hist_toobig = 0;
int GAP_HIST_DIVISOR;

static void wait_for_all_threads (enum thrd_status_t new_status)
{
  int cnt = 0;
  int ready;
  int i;

  do {
    if (cnt++ > MAX_SPIN) {
      printf ("Hrm: waited too long for threads to reach state %d\n",
	      new_status);
      exit (-1);
    }
    usleep (SMALL_SLEEP);
    ready = 1;
    for (i=0; i<num_threads; i++) {
      if (thrd_data[i].status != new_status) {
	ready = 0;
      }
    }
  } while (!ready);
}

static void set_all_threads (enum thrd_status_t new_status)
{
  int i;

  for (i=0; i<num_threads; i++) {
    thrd_data[i].status = new_status;
  }
}

static void flush_records (void)
{
  int i, rec;

  for (i=0; i<num_threads; i++) {
    rec = thrd_data[i].reserved_record;
    if (rec < 0) {
      continue;
    }
    trace_recs[rec].id = i;
    trace_recs[rec].start = thrd_data[i].exec_start;
    trace_recs[rec].end = thrd_data[i].last_exec;
  }
}
  
pid_t hourglass_pid;

int main (int argc, char *argv[])
{
  int i, retval;

  hourglass_pid = getpid ();

  gap_us = 5.0; // FIXME: this is a platform dependent hack
  polling_overhead_us = 0.2; // FIXME: this is a platform dependent hack

  duration = 10000000;  // the default

  parse_args (argc, argv);
  
  if (num_threads == 0) {
    printf ("Hrm: internal error\n");
    exit (-1);
  }

  if (TSC_MHZ == -1) {
    // time conversion functions won't work until this runs
    TSC_MHZ = find_mhz ();
  }

  if (generate_gap_hist) {
    GAP_HIST_DIVISOR = us_to_cycle (gap_us) / NUM_GAP_HIST_BUCKETS;
    // random
    GAP_HIST_DIVISOR = mymax (GAP_HIST_DIVISOR, 10);
  }

  printf ("Hourglass %s : %d thread%s; %f seconds; %f MHz timestamp counter\n",
	  VERSION, 
	  num_threads, 
	  (num_threads == 1) ? "" : "s",
	  us_to_sec ((double)duration),
	  TSC_MHZ);
  
  // system ("uname -a");
  // system ("date");

  if (tracing) {
    gap = us_to_cycle (gap_us);
    polling_overhead = us_to_cycle (polling_overhead_us);
    printf ("max gap is %lld cycles\n", gap);
  }

  lock_pages ();
  set_priority (MAIN_PRI);

#ifdef WITH_RTCTIMERS
  if (need_rtc_timers) {
    rtc_timer_funcs.init_timers ();
  }
#endif

#ifdef WIN32
  if (need_mm_timers) {
    mm_timer_funcs.init_timers ();
  }
#endif

  printf ("this test will last for %f seconds\n", 
	  us_to_sec ((double)duration));

  set_all_threads (INIT);
  
  // sync up with a clock interrupt
  usleep (15000); 

  for (i=0; i<num_threads; i++) {
    retval = pthread_create (&thrd_data[i].thrd,
			     NULL,
			     thrd_func,
			     (void *)i);
    if (retval != 0) {
      printf ("Hrm: error creating thread %d\n", i);
      exit (-1);
    }

    if (thrd_data[i].f_time != -1) {
      retval = pthread_create (&thrd_data[i].monitor_thrd,
			       NULL,
			       monitor_thrd_func,
			       (void *)i);
      if (retval != 0) {
	printf ("Hrm: error creating monitor thread %d\n", i);
	exit (-1);
      }
    }
  }

  set_all_threads (GO);

  usleep (INITIAL_DELAY);

  start_time = rdtsc();

  trace_start = rdtsc ();
  usleep (duration);
  trace_end = rdtsc ();

  /*
   * threads that were already stopped by their monitor threads
   * will get to DONE on their own
   */
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].status == GO) {
      thrd_data[i].status = STOP;
    }
  }

  wait_for_all_threads (DONE);
  end_time = rdtsc();
  if (tracing) {
    flush_records ();
  }
  set_priority (NORMAL);

#ifdef WIN32
  if (need_mm_timers) {
    mm_timer_funcs.deinit_timers ();
  }
#endif

  print_all_results ();

  return 0;
} 
