/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

int NUM_TRACE_RECORDS = DEF_NUM_TRACE_RECORDS;

static void usage (void)
{
  printf ("This is Hourglass %s\n", VERSION);
  printf ("\n");
  printf ("usage: hourglass <options>\n");
  printf ("\n");
  printf ("global options:\n");
  printf ("  -a -- subsequent per-thread options apply to\n");
  printf ("     all threads; this remains in effect until\n");
  printf ("     the next '-t' option is encountered\n");
  printf ("  -c -- print raw execution trace (in cycles and\n");
  printf ("     not based at zero) in addition to standard\n");
  printf ("     trace (which starts at zero and is in ms)\n");
  printf ("  -d <duration of experiment> (default is 10s)\n");
  printf ("  -D <level> -- debug level, 0 == quiet (default)\n");
  printf ("    10 == loudest\n");
  printf ("  -e -- number of records in execution trace;\n");
  printf ("     default is %d (about %d MB)\n",
	  DEF_NUM_TRACE_RECORDS, 
	  (sizeof (struct trace_rec_t) * DEF_NUM_TRACE_RECORDS) / (1024*1024));
  printf ("  -g <gap> -- set the size of the smallest gap in\n");
  printf ("     thread execution that will be recorded as a\n");
  printf ("     gap (default %fus)\n", gap_us);
  printf ("  -gh -- print a histogram of the sizes of gaps in\n");
  printf ("     thread execution that are smaller than the gap\n");
  printf ("     threshold\n");
  printf ("  -m -- instead of using built-in portable timestamp\n");
  printf ("     counter frequency detection, use a non-portable source\n");
  printf ("     (if available) (WARNING -- broken on PPC Linux!)\n");
  printf ("  -M <MHz> -- directly specify MHz of the timestamp counter\n");
  printf ("  -n <number of threads to create> -- this is\n");
  printf ("     the only mandatory option\n");
  printf ("  -po <ovhd> -- set the overhead of the polling loop\n");
  printf ("     (default %fus)\n", polling_overhead_us);
  printf ("  -t <thread number for subsequent options to\n");
  printf ("     affect in the range 0..n-1>\n");
  printf ("  -x -- suppress printing of standard execution\n");
  printf ("     trace\n");
  printf ("\n");
  printf ("per-thread options:\n");
  printf ("  -f <time> -- terminate this thread <time>\n");
  printf ("    units into the Hourglass run\n");
  printf ("  -i <timer>\n");
  printf ("    NATIVE -- native Unix or Windows timers\n");
  printf ("      (default)\n");
  printf ("    HR -- high resolution timers (if available:\n");
  printf ("      requires a patched kernel)\n");
  printf ("    RTC -- timers based on Linux real-time\n");
  printf ("      clock (if available)\n");
  printf ("    MM -- multimedia timers (Win32 only)\n");
  printf ("  -p <priority> -- priority is one of:\n");
  printf ("    IDLE, LOW, NORMAL, HIGH, HIGHEST\n");
  printf ("    (timesharing -- default is NORMAL)\n");
  printf ("    RTLOW, RTMED, RTHIGH (real-time)\n");
  printf ("  -rh <amount> <period> -- request a hard\n");
  printf ("    CPU reservation\n");
  printf ("  -rs <amount> <period> -- request a soft\n");
  printf ("    CPU reservation\n");
  printf ("  -s <time> -- start running this thread <time>\n");
  printf ("    units into the Hourglass run\n");
  printf ("  -w <workload>\n");
  printf ("    CPU -- cpu-bound execution trace (default)\n");
  printf ("    CPU_SCAN <size> -- cpu-bound and also scans\n");
  printf ("      an array that is size KB long\n");
  printf ("    CPU_YIELD <amount> -- yield the processor\n");
  printf ("      after getting <amount> of CPU time\n");
  printf ("    CPU_SCAN_YIELD <size> <amount> -- combine\n");
  printf ("      scanning and yielding behaviors\n");
  printf ("    PERIODIC <amount> <period> -- periodic\n");
  printf ("      task, make execution trace\n");
  printf ("    CPU_PERIODIC <amount> <period> -- cpu-bound\n");
  printf ("      periodic (see documentation)\n");
  printf ("    LAT <period> -- dispatch latency test\n");
  printf ("\n");
  printf ("Times are self-describing; valid examples are:\n");   
  printf ("3m, 1.5s, 1020ms, 87.0us.\n");
  printf ("\n");
  exit (-1);
}

static void init_thrd_data (void)
{
  int i;

  if (thrd_data) {
    printf ("oops: number of threads specified twice\n");
    exit (-1);
  }
  
  thrd_data = (struct thrd_data_t *) 
    xmalloc (num_threads * sizeof (struct thrd_data_t));

  for (i=0; i<num_threads; i++) {
    // initialize vars
    thrd_data[i].exec_start = 0;
    thrd_data[i].last_exec = 0;
    thrd_data[i].reserved_record = -1;
    thrd_data[i].lat_slots = NULL;
    thrd_data[i].res_period = thrd_data[i].res_amount = -1;
    thrd_data[i].app_period = thrd_data[i].app_amount = -1;
    thrd_data[i].memsize = (unsigned int)-1;
    thrd_data[i].scanmem = NULL;

    // set defaults
    thrd_data[i].s_time = thrd_data[i].f_time = -1;
    thrd_data[i].timers = &unix_timer_funcs;
    thrd_data[i].log = 0;
    thrd_data[i].pri = NORMAL;
    thrd_data[i].work = CPU;
  }
}

static int pos;
static int argc;
static char **argv;

static int expect_int (void)
{
  int i;
  char *c;

  if (pos >= argc) {
      printf ("oops: integer expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  i = strtol (argv[pos], &c, 10);
  if (*c != '\0') {
    printf ("oops: integer expected at cmd line position %d\n",
	    pos);
    exit (-1);
  }

  return i;
}

#if 0
static double expect_double (void)
{
  double d;
  char *c;

  if (pos >= argc) {
      printf ("oops: float expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  d = strtod (argv[pos], &c);
  if (*c != '\0') {
    printf ("oops: float expected at cmd line position %d\n",
	    pos);
    exit (-1);
  }

  return d;
}
#endif

static us_time expect_time (void)
{
  int len;
  double t;
  char *c;

  if (pos >= argc) {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  len = strlen (argv[pos]);

  // microseconds
  if (len >= 3 && strcmp (&argv[pos][len-2], "us") == 0) {
    argv[pos][len-2] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t;
  }

  // milliseconds
  if (len >= 3 && strcmp (&argv[pos][len-2], "ms") == 0) {
    argv[pos][len-2] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t * 1000;
  }

  // minutes
  if (len >= 2 && strcmp (&argv[pos][len-1], "m") == 0) {
    argv[pos][len-1] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t * 1000000 * 60;
  }

  // seconds - this needs to be last
  if (len >= 2 && strcmp (&argv[pos][len-1], "s") == 0) {
    argv[pos][len-1] = '\0';
    t = strtod (argv[pos], &c);
    if (*c != '\0') {
      printf ("oops: time value expected at cmd line position %d\n",
	      pos);
      exit (-1);
    }
    return (us_time) t * 1000000;
  }

  printf ("oops: time value expected at cmd line position %d\n",
	  pos);
  exit (-1);
}

static enum thrd_pri_t expect_pri (void)
{
  if (pos >= argc) {
      printf ("oops: priority expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  if (strcmp (argv[pos], "IDLE") == 0) {
    return IDLE;
  } else if (strcmp (argv[pos], "LOW") == 0) {
    return LOW;
  } else if (strcmp (argv[pos], "NORMAL") == 0) {
    return NORMAL;
  } else if (strcmp (argv[pos], "HIGH") == 0) {
    return HIGH;
  } else if (strcmp (argv[pos], "HIGHEST") == 0) {
    return HIGHEST;
  } else if (strcmp (argv[pos], "RTLOW") == 0) {
    return RTLOW;
  } else if (strcmp (argv[pos], "RTMED") == 0) {
    return RTMED;
  } else if (strcmp (argv[pos], "RTHIGH") == 0) {
    return RTHIGH;
  }

  printf ("oops: priority expected at cmd line position %d\n",
	  pos);
  exit (-1);
}

static const char *print_pri (enum thrd_pri_t pri) 
{
  switch (pri) {
  case IDLE:
    return "IDLE";
  case LOW:
    return "LOW";
  case NORMAL:
    return "NORMAL";
  case HIGH:
    return "HIGH";
  case HIGHEST:
    return "HIGHEST";
  case RTLOW:
    return "RTLOW";
  case RTMED:
    return "RTMED";
  case RTHIGH:
    return "RTHIGH";
  case TIMER_PRI:
    return "TIMER_PRI";
  case MAIN_PRI:
    return "MAIN_PRI";
  default:
    return "unknown priority!";
  }
}

static enum thrd_work_t expect_work (void)
{
  if (pos >= argc) {
      printf ("oops: workload expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  if (strcmp (argv[pos], "CPU") == 0) {
    return CPU;
  } else if (strcmp (argv[pos], "PERIODIC") == 0) {
    return PERIODIC;
  } else if (strcmp (argv[pos], "CPU_PERIODIC") == 0) {
    return CPU_PERIODIC;
  } else if (strcmp (argv[pos], "CPU_SCAN") == 0) {
    return CPU_SCAN;
  } else if (strcmp (argv[pos], "CPU_SCAN_YIELD") == 0) {
    return CPU_SCAN_YIELD;
  } else if (strcmp (argv[pos], "CPU_YIELD") == 0) {
    return CPU_YIELD;
  } else if (strcmp (argv[pos], "LAT") == 0) {
    return LAT;
  }

  printf ("oops: workload expected at cmd line position %d\n",
	  pos);
  exit (-1);
}

static const char *print_work (enum thrd_work_t pri) 
{
  switch (pri) {
  case CPU:
    return "CPU";
  case LAT:
    return "LAT";
  case PERIODIC:
    return "PERIODIC";
  case CPU_PERIODIC:
    return "CPU_PERIODIC";
  case CPU_SCAN:
    return "CPU_SCAN";
  case CPU_SCAN_YIELD:
    return "CPU_YIELD";
  case CPU_YIELD:
    return "CPU_YIELD";
  default:
    assert (0);
  }
  return NULL;
}

static const struct timer_funcs * expect_timer (void)
{
  if (pos >= argc) {
      printf ("oops: priority expected at cmd line position %d\n",
	      pos);
      exit (-1);
  }

  if (strcmp (argv[pos], "NATIVE") == 0) {
    return &unix_timer_funcs;
  } else if (strcmp (argv[pos], "HR") == 0) {
#ifdef WITH_HRTIMERS
    return &hr_timer_funcs;
#else
    printf ("oops: high-res timers not supported in this build\n");
    exit (-1);
#endif
  } else if (strcmp (argv[pos], "RTC") == 0) {
#ifdef WITH_RTCTIMERS
    return &rtc_timer_funcs;
#else
    printf ("oops: real-time clock timers not supported in this build\n");
    exit (-1);
#endif
  } else if (strcmp (argv[pos], "MM") == 0) {
#ifdef WIN32
    return &mm_timer_funcs;
#else
    printf ("oops: multimedia timers not supported in this build\n");
    exit (-1);
#endif
  }

  printf ("oops: timer expected at cmd line position %d\n",
	  pos);
  exit (-1);
}

static const char *print_timer (const struct timer_funcs *timer) 
{
  if (timer == &unix_timer_funcs) {
    return "Unix";

#ifdef WITH_RTCTIMERS
  } else if (timer == &rtc_timer_funcs) {
    return "RTC";
#endif

#ifdef WIN32
  } else if (timer == &mm_timer_funcs) {
    return "multimedia";
#endif

#ifdef WITH_HRTIMERS
  } else if (timer == &hr_timer_funcs) {
    return "high-res";
#endif

  } else {
    return "unknown timer!";
  }
}

static void set_thrd_pri (int id, enum thrd_pri_t pri)
{
  thrd_data[id].pri = pri;
  printf ("thread %d will have priority %s\n",
	  id, print_pri (pri));
}

static void set_thrd_timer (int id, const struct timer_funcs *timer)
{
  thrd_data[id].timers = timer;
  printf ("thread %d will use %s timers\n",
	  id, print_timer (timer));
}

static void init_lat_slots (int id) 
{
  thrd_data[id].lat_slots = (cycle_time *) malloc (sizeof (cycle_time) * LAT_SLOTS);
}

static void set_thrd_work (int id, enum thrd_work_t work)
{
  thrd_data[id].work = work;
  printf ("thread %d will run workload %s\n",
	  id, print_work (work));
}

#if (defined WITH_CMU_RK || defined WITH_TIMESYS_RK || defined WITH_HLS)
static void set_thrd_res_pa (int id, int amount, int period, int soft)
{
  thrd_data[id].res_amount = amount;
  thrd_data[id].res_period = period;
  thrd_data[id].res_soft = soft;
  printf ("thread %d will have %s res of amount %d and period %d\n",
	  id, 
	  (soft) ? "soft" : "hard",
	  amount,
	  period);
}
#endif

static void set_thrd_period (int id, int period)
{
  thrd_data[id].app_period = period;
  printf ("thread %d will have period %d\n", id, period);
}

static void set_thrd_amount (int id, int amount)
{
  thrd_data[id].app_amount = amount;
  printf ("thread %d will have amount %d\n", id, amount);
}

static void set_thrd_memsize (int id, unsigned int memsize)
{
  thrd_data[id].memsize = memsize * 1024;
  if (memsize != 0) {
    thrd_data[id].scanmem = (int *) xmalloc ((unsigned int)(1024 * memsize));
  } else {
    thrd_data[id].scanmem = 0;
  }
  printf ("thread %d will scan %d KB (%d bytes, %d ints)\n",
	  id, memsize, thrd_data[id].memsize, 1024*memsize/sizeof(int));
}

static void set_start_time (int id, us_time s_time)
{
  thrd_data[id].s_time = s_time;
  printf ("thread %d will start %f s into the run\n", id, us_to_sec(s_time));
}

static void set_finish_time (int id, us_time f_time)
{
  thrd_data[id].f_time = f_time;
  printf ("thread %d will terminate %f s into the run\n", id, us_to_sec(f_time));
}

int print_cycles = 0;

void parse_args (int passed_argc, char *passed_argv[])
{
  int i;
  int cur_thread = -1;
  int all_threads = 0;

  argc = passed_argc;
  argv = passed_argv;

  if ((argc < 2) ||
      (strcmp (argv[1], "-h") == 0) ||
      (strcmp (argv[1], "-?") == 0)) {
    usage ();
  }

  pos = 1;
  while (pos < argc) {

    if (strcmp (argv[pos], "-n") == 0) {
      pos++;
      num_threads = expect_int ();
      if (num_threads > 50) {
	printf ("oops: max of 50 threads: creating too many has been observed\n");
	printf ("to lock up the Linux kernel.\n");
	exit (-1);
      }
      pos++;
      // printf ("%d threads\n", num_threads);
      init_thrd_data ();
      continue;
    }

    if (strcmp (argv[pos], "-d") == 0) {
      pos++;
      duration = (unsigned int) expect_time ();
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-g") == 0) {
      pos++;
      gap_us = expect_time ();
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-po") == 0) {
      pos++;
      polling_overhead_us = expect_time ();
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-a") == 0) {
      cur_thread = -1;
      all_threads = 1;
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-x") == 0) {
      suppress_trace = 1;
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-c") == 0) {
      print_cycles = 1;
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-t") == 0) {
      pos++;
      if (!thrd_data) {
	printf ("oops: option -t may only appear after -n\n");
	exit (-1);
      }
      cur_thread = expect_int ();
      all_threads = 0;
      pos++;
      if (cur_thread < 0 || cur_thread >= num_threads) {
	printf ("oops: thread number %d out of range\n", cur_thread);
	exit (-1);
      }
      continue;
    }

    if (strcmp (argv[pos], "-e") == 0) {
      pos++;
      NUM_TRACE_RECORDS = expect_int ();
      pos++;
      if (NUM_TRACE_RECORDS < 0) {
	printf ("oops: bad number of trace records %d\n", NUM_TRACE_RECORDS);
	exit (-1);
      }
      continue;
    }

    if (strcmp (argv[pos], "-gh") == 0) {
      pos++;
      generate_gap_hist = TRUE;
      continue;
    }

    if (strcmp (argv[pos], "-m") == 0) {
      pos++;
      DBGPrint (3, ("trying to locate MHz info in /proc/cpuinfo\n"));
      {
	FILE *f = fopen ("/proc/cpuinfo", "r");
	int res;
	if (!f) {
	  printf ("oops: can't open /proc/cpuinfo -- MHz calculation\n");
	  printf ("will use the internal routine\n");
	  continue;
	}

	do {
	  float speed;

	  res = fscanf (f, "cpu MHz : %f", &speed);
	  printf ("foo %d\n", res);
	  if (res == 1) {
	    if (speed > 0.0) {
	      TSC_MHZ = speed;
	      DBGPrint (3, ("/proc/cpuinfo says %f MHz\n", TSC_MHZ));
	    }
	  }
	} while (!feof (f));

	fclose (f);
      }
      continue;
    }

    if (strcmp (argv[pos], "-M") == 0) {
      pos++;
      TSC_MHZ = (double) expect_int ();
      pos++;
      if (NUM_TRACE_RECORDS < 0) {
	printf ("oops: bad TSC_MHZ %d\n", NUM_TRACE_RECORDS);
	exit (-1);
      }
      continue;
    }

    if (strcmp (argv[pos], "-D") == 0) {
      pos++;
      _DBG_LEVEL_ = expect_int ();
      pos++;
      continue;
    }

    if (strcmp (argv[pos], "-p") == 0) {
      enum thrd_pri_t pri;
      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-p) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      pri = expect_pri ();
      pos++;
      
      if (all_threads) {
	int j;
	for (j=0; j<num_threads; j++) {
	  set_thrd_pri (j, pri);
	}
      } else {
	set_thrd_pri (cur_thread, pri);
      }
      continue;
    }

    if (strcmp (argv[pos], "-s") == 0) {
      us_time s_time;
      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-s) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      s_time = (unsigned int) expect_time ();
      pos++;

      if (all_threads) {
	int j;
	for (j=0; j<num_threads; j++) {
	  set_start_time (j, s_time);
	}
      } else {
	set_start_time (cur_thread, s_time);
      }

      continue;
    }

    if (strcmp (argv[pos], "-f") == 0) {
      us_time f_time;
      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-f) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      f_time = (unsigned int) expect_time ();
      pos++;

      if (all_threads) {
	int j;
	for (j=0; j<num_threads; j++) {
	  set_finish_time (j, f_time);
	}
      } else {
	set_finish_time (cur_thread, f_time);
      }

      continue;
    }

    if (strcmp (argv[pos], "-i") == 0) {
      const struct timer_funcs *timer;

      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-i) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      timer = expect_timer ();
      pos++;

      if (all_threads) {
	int j;
	for (j=0; j<num_threads; j++) {
	  set_thrd_timer (j, timer);
	}
      } else {
	set_thrd_timer (cur_thread, timer);
      }

      continue;
    }

    if (strcmp (argv[pos], "-w") == 0) {
      enum thrd_work_t work;
      int period = -1, amount = -1;
      unsigned int memsize = (unsigned int)-1;

      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-i) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
      pos++;
      work = expect_work ();
      pos++;

      if (work == PERIODIC || work == CPU_PERIODIC) {
	amount = (int) expect_time ();
	pos++;
	period = (int) expect_time ();
	pos++;
      } else if (work == LAT) {
	period = (int) expect_time ();
	pos++;
      } else if (work == CPU_YIELD) {
	amount = (int) expect_time ();
	pos++;
      } else if (work == CPU_SCAN) {
	memsize = expect_int ();
	pos++;
      } else if (work == CPU_SCAN_YIELD) {
	memsize = expect_int ();
	pos++;
	amount = (int) expect_time ();
	pos++;
      }

      if (all_threads) {
	int j;
	for (j=0; j<num_threads; j++) {
	  set_thrd_work (j, work);
	  switch (work) {
	  case LAT:
	    init_lat_slots (j);
	    set_thrd_period (j, period);
	    break;
	  case CPU_YIELD:
	    set_thrd_amount (j, amount);
	    break;
	  case PERIODIC:
	  case CPU_PERIODIC:
	    set_thrd_amount (j, amount);
	    set_thrd_period (j, period);
	    break;
	  case CPU:
	    break;
	  case CPU_SCAN:
	    set_thrd_memsize (j, memsize);
	    break;
	  case CPU_SCAN_YIELD:
	    set_thrd_amount (j, amount);
	    set_thrd_memsize (j, memsize);
	    break;
	  default:
	    printf ("oops: internal error!\n");
	    exit (-1);
	  }
	}
      } else {
	set_thrd_work (cur_thread, work);
	switch (work) {
	case LAT:
	  init_lat_slots (cur_thread);
	  set_thrd_period (cur_thread, period);
	  break;
	case CPU_YIELD:
	  set_thrd_amount (cur_thread, amount);
	  break;
	case PERIODIC:
	case CPU_PERIODIC:
	  set_thrd_amount (cur_thread, amount);
	  set_thrd_period (cur_thread, period);
	  break;
	case CPU:
	  break;
	case CPU_SCAN:
	  set_thrd_memsize (cur_thread, memsize);
	  break;
	case CPU_SCAN_YIELD:
	  set_thrd_amount (cur_thread, amount);
	  set_thrd_memsize (cur_thread, memsize);
	  break;
	default:
	  printf ("oops: internal error!\n");
	  exit (-1);
	}
      }

      continue;
    }

    if (strcmp (argv[pos], "-rs") == 0 ||
	strcmp (argv[pos], "-rh") == 0) {
      if (cur_thread == -1 && !all_threads) {
	printf ("oops: option %d (-r) is only meaningful for a thread\n",
		pos);
	exit (-1);
      }
#if (defined WITH_HLS || defined WITH_CMU_RK || defined WITH_TIMESYS_RK)
      {
	int period, amount, soft;

	if (strcmp (argv[pos], "-rs") == 0) {
	  soft = 1;
	} else {
	  soft = 0;
	}

	pos++;
	amount = expect_time ();
	pos++;
	period = expect_time ();
	pos++;
	
	if (all_threads) {
	  int j;
	  for (j=0; j<num_threads; j++) {
	    set_thrd_res_pa (j, amount, period, soft);
	  }
	} else {
	  set_thrd_res_pa (cur_thread, amount, period, soft);
	}
      }

      continue;
#else
      printf ("oops: CPU reservations are not supported by this build\n");
      exit (-1);
#endif
    }

    printf ("oops: unexpected arg '%s' at position %d\n", argv[pos], pos);
    exit (-1);
  }

  if (num_threads == 0) {
    printf ("oops: need to specify number of threads to create\n");
    exit (-1);
  }

#ifdef WITH_RTCTIMERS
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].timers == &rtc_timer_funcs) {
      need_rtc_timers = 1;
    }
  }
#endif

#ifdef WIN32
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].timers == &mm_timer_funcs) {
      need_mm_timers = 1;
    }
  }
#endif

  tracing = 0;
  for (i=0; i<num_threads; i++) {
    if (thrd_data[i].work == CPU ||
	thrd_data[i].work == PERIODIC ||
	thrd_data[i].work == CPU_PERIODIC ||
	thrd_data[i].work == CPU_SCAN ||
	thrd_data[i].work == CPU_SCAN_YIELD ||
	thrd_data[i].work == CPU_YIELD) {
      tracing = 1;
    }
  }

  if (tracing) {
    trace_recs = (struct trace_rec_t *)
      xmalloc (sizeof (struct trace_rec_t) * NUM_TRACE_RECORDS);

    printf ("%f MB allocated for trace records\n",
	    (sizeof (struct trace_rec_t) * NUM_TRACE_RECORDS) / 1024.0 / 1024.0);

    if (generate_gap_hist) {
      gap_hist = (int *) 
	xmalloc (sizeof (int) * NUM_GAP_HIST_BUCKETS);    
      for (i=0; i<NUM_GAP_HIST_BUCKETS; i++) {
	gap_hist[i] = 0;
      }
    }

  } else {
    trace_recs = NULL;
  }
  current_record = 0;
}
