/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include "hourglass.h"

// FIXME -- have autoconf take care of this

// Linux 2.4
#include <linux/rtc.h>

// Linux 2.2
// #include <linux/mc146818rtc.h>

/*
 * timers based on RTC
 *
 * cheezy because:
 *   - one timer per thread
 *   - setting will bomb if already set
 *   - to avoid kernel mods, wakeups are done by a thread, causing
 *     many extra context switches
 *
 * good because:
 *   - get high-res wakeups with standard kernel
 */

/* 
 * set RTC to 4 KHZ, meaning that ideally, wakeups will happen no more
 * than 244us late 
 */
#define CLOCK_HZ 4096

static int RTCLOCKFD;
static int RTC_TIMER_INIT = 0;

struct rtc_timer {
  int set;
  us_time expire;
  pthread_mutex_t mut;
  pthread_cond_t cond;
};

static struct rtc_timer *rtc_timers;
static volatile int running;

static void rtc_timed_sleep (int id, us_time t)
{
  ASSERT (RTC_TIMER_INIT);

  pthread_mutex_lock (&rtc_timers[id].mut);
  ASSERT (!rtc_timers[id].set);
  rtc_timers[id].set = 1;

  if (t < 0) {
    // negative time is relative sleep
    us_time now = cycle_to_us (rdtsc());
    rtc_timers[id].expire = now - t;
  } else {
    // positive time is absolute
    rtc_timers[id].expire = t;
  }

  ASSERT (rtc_timers[id].set);

  pthread_cond_wait (&rtc_timers[id].cond,
		     &rtc_timers[id].mut);

  ASSERT (!rtc_timers[id].set);

  pthread_mutex_unlock (&rtc_timers[id].mut);
}

static void *timer_thrd_func (void *arg)
{
  int retval;
  unsigned long stuff;

  set_priority (TIMER_PRI);

  while (running == 1) {
    int i;
    us_time now = cycle_to_us (rdtsc ());

    for (i=0; i<num_threads; i++) {
      if (rtc_timers[i].set && 
	  now >= rtc_timers[i].expire) {
	pthread_mutex_lock (&rtc_timers[i].mut);
	rtc_timers[i].set = 0;
	pthread_cond_signal (&rtc_timers[i].cond);
	pthread_mutex_unlock (&rtc_timers[i].mut);
      }
    }

    retval = read (RTCLOCKFD, &stuff, sizeof (unsigned long));
    if (retval == -1) {
      perror("read");
      exit(errno);
    }
  }
  running++;
  return NULL;
}

static void rtc_init_timers (void)
{
  int retval, i;
  pthread_attr_t attr;
  pthread_t thrd;

  ASSERT (!RTC_TIMER_INIT);
  RTC_TIMER_INIT = 1;
    
  RTCLOCKFD = open ("/dev/rtc", O_RDONLY);
  
  if (RTCLOCKFD ==  -1) {
    perror ("/dev/rtc");
    exit (errno);
  }
  
  /* Set periodic interrupt frequency */
  retval = ioctl (RTCLOCKFD, RTC_IRQP_SET, CLOCK_HZ);
  if (retval == -1) {
    perror ("ioctl");
    exit (errno);
  }

  /* Enable periodic interrupts */
  retval = ioctl (RTCLOCKFD, RTC_PIE_ON, 0);
  if (retval == -1) {
    perror("ioctl");
    exit(errno);
  }

  rtc_timers = (struct rtc_timer *) 
    xmalloc (num_threads * sizeof (struct rtc_timer));

  for (i=0; i<num_threads; i++) {
    rtc_timers[i].set = 0;
    pthread_mutex_init (&rtc_timers[i].mut, NULL);
    pthread_cond_init (&rtc_timers[i].cond, NULL);
  }

  running = 1;

  retval = pthread_create (&thrd,
			   NULL,
			   timer_thrd_func,
			   NULL);
  if (retval != 0) {
    printf ("Hrm: error creating thread %d\n", i);
    exit (-1);
  }
}

static void rtc_deinit_timers (void)
{
  int cnt = 0;

  ASSERT (RTC_TIMER_INIT);
  RTC_TIMER_INIT = 0;

  running++;
  do {
    if (cnt++ > MAX_SPIN) {
      printf ("Hrm: waited too long for timer thread to exit\n");
      exit (-1);
    }
    usleep (SMALL_SLEEP);
  } while (running == 2);

  close (RTCLOCKFD);
}

const struct timer_funcs rtc_timer_funcs = {
  rtc_timed_sleep,
  rtc_init_timers,
  rtc_deinit_timers
};
